#parse (remember to do static methods, unless you're changing the formatting)

class StartParse(object): #really hope this doesn't take too long to run...
	@staticmethod
	def ready(response):
		ready_data = dict(response["d"])
		ready_data.pop("merged_members")
		#parse relationships
		ready_data["relationships"] = {i["id"]:i for i in response["d"]["relationships"]}
		#parse private channels
		ready_data["private_channels"] = {j["id"]:j for j in response["d"]["private_channels"]}
		#add activities key to user settings
		ready_data["user_settings"]["activities"] = {}
		#parse guilds
		guilds = response["d"]["guilds"]
		ready_data["guilds"] = {k["id"]:k for k in guilds}
		for personal_role, guild in zip(response["d"]["merged_members"], guilds):
			#take care of personal role/nick
			ready_data["guilds"][guild["id"]]["my_data"] = personal_role
			#take care of emojis
			ready_data["guilds"][guild["id"]]["emojis"] = {l["id"]:l for l in guild["emojis"]}
			#take care of roles
			ready_data["guilds"][guild["id"]]["roles"] = {m["id"]:m for m in guild["roles"]}
			#take care of channels
			ready_data["guilds"][guild["id"]]["channels"] = {n["id"]:n for n in guild["channels"]}
			#take care of members
			ready_data["guilds"][guild["id"]]["members"] = {}
		return ready_data

	@staticmethod
	def ready_supplemental(response):
		ready_supp_data = dict(response["d"])
		ready_supp_data["online_friends"] =  {o["user_id"]:o for o in response["d"]["merged_presences"]["friends"]}
		ready_supp_data.pop("guilds")
		ready_supp_data["voice_states"] = {p["id"]:p["voice_states"] for p in response["d"]["guilds"]} #id is the guild_id
		return ready_supp_data


from ..RESTapiwrap import *

class Stickers(object):
	def __init__(self, discord, s, log): #s is the requests session object
		self.discord = discord
		self.s = s
		self.log = log

	def getStickers(self, directoryID="758482250722574376", store_listings=False, locale="en-US"):
		store_listings = str(store_listings).lower()
		url = self.discord+"sticker-packs/directory-v2/"+directoryID+"?with_store_listings="+store_listings+"&locale="+locale
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	def getStickerFile(self, stickerID, stickerAsset): #this is an apng
		url = "https://media.discordapp.net/stickers/"+stickerID+"/"+stickerAsset+".png"
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	def getStickerJson(self, stickerID, stickerAsset):
		url = "https://discord.com/stickers/"+stickerID+"/"+stickerAsset+".json"
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	def getStickerPack(self, stickerPackID):
		url = self.discord+"sticker-packs/"+stickerPackID
		return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

	


import copy #only used if header modification is used
import json
import inspect
import brotli #just in case

def brdecompress(payload):
	try:
		data = brotli.decompress(payload)
		return data
	except:
		return payload

def editedReqSession(reqsession, headerModifications):
	editedSession = copy.deepcopy(reqsession)
	if "add" in headerModifications:
		editedSession.headers.update(headerModifications["update"])
	if "remove" in headerModifications:
		for header in headerModifications["remove"]:
			if header in editedSession.headers:
				del editedSession.headers[header]
	return editedSession

class LogLevel:
	INFO = '\033[94m'
	OK = '\033[92m'
	WARNING = '\033[93m'
	DEFAULT = '\033[m'

#The wrap that all http request api functions in discum use
class Wrapper:
	@staticmethod
	def sendRequest(reqsession, method, url, body=None, files=None, headerModifications={}, timeout=None, log=True): #headerModifications = {"add":{}, "remove":[]}
		if hasattr(reqsession, method): #just checks if post, get, whatever is a valid requests method
			stack = inspect.stack()
			function_name = "({}->{})".format(str(stack[1][0].f_locals['self']).split(' ')[0], stack[1][3])
			if log: #(sent) log msg, method and url
				sentMsg = '{} -> {}'.format(method.title(), url)
				print('{} [+] {} {}'.format(LogLevel.INFO, function_name, sentMsg))
				print(LogLevel.DEFAULT)
			data = {} #now onto the body (if exists)
			if body != None:
				if isinstance(body, dict):
					data = {'data': json.dumps(body)}
				else:
					data = {'data': body}
				if log:
					bodyMsg = str(body)
					print('{} [+] {} {}'.format(LogLevel.INFO, function_name, bodyMsg))
					print(LogLevel.DEFAULT)
			if files != None: #files...this isnt actually used yet in discum
				data['files'] = files
				if log:
					print('{} [+] {} {}'.format(LogLevel.INFO, function_name, "<file data>"))
					print(LogLevel.DEFAULT)
			prepReqSession = reqsession
			if headerModifications not in ({}, None): #header modifications, like endpoints that don't need auth or superproperties or stuff like that
				prepReqSession = editedReqSession(prepReqSession, headerModifications)
			if timeout == None:
				response = getattr(prepReqSession, method)(url=url, **data) #where the actual request happens
			else:
				data['timeout'] = timeout
				try:
					response = getattr(prepReqSession, method)(url=url, **data)
					return
				except:
					return
			if response.headers.get('Content-Encoding') == "br": #decompression; gzip/deflate is automatically handled by requests module
				response._content = brdecompress(response.content)
			if log: #(received) log message, response
				receivedMsg = 'Response <- {}'.format(response.text)
				print('{} [+] {} {}'.format(LogLevel.OK, function_name, receivedMsg))
				print(LogLevel.DEFAULT)
			return response
		else:
			print('Invalid request method.')



from ..RESTapiwrap import *

#This file includes apis that run when your client starts but idk where to organize them
#Maybe they'll get organized eventually, idk
class Other:
    def __init__(self, s, discordurl, log):
        self.s = s
        self.discord = discordurl
        self.log = log

    def getGatewayUrl(self):
        url = self.discord + "gateway"
        return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

    def getDiscordStatus(self):
        url = "https://status.discord.com/api/v2/scheduled-maintenances/upcoming.json"
        return Wrapper.sendRequest(self.s, 'get', url, headerModifications={"remove":["Authorization", "X-Super-Properties"]}, log=self.log)

    def getDetectables(self):
        url = self.discord + "applications/detectable"
        return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

    def getOauth2Tokens(self):
        url = self.discord + "oauth2/tokens"
        return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

    def getVersionStableHash(self, underscore=None):
        if isinstance(underscore, int):
            underscore = str(underscore)
        url = "https://discord.com/assets/version.stable.json"
        if isinstance(underscore, str):
            url += "?_="+underscore
        return Wrapper.sendRequest(self.s, 'get', url, headerModifications={"remove":["Authorization", "X-Super-Properties"]}, log=self.log)


#handles TOTP (Time-based One-Time Password) code calculation for Discord. Discord's totp refreshes every 30 seconds.
#implementation is the same as https://github.com/pyauth/pyotp/ (most of the functions are just copy-pasted)

import time, datetime
import hmac
import base64
import hashlib

class TOTP:
    def __init__(self, secret):
        self.secret = secret

    def byte_secret(self):
        missing_padding = len(self.secret) % 8
        if missing_padding != 0:
            self.secret += '=' * (8 - missing_padding)
        return base64.b32decode(self.secret, casefold=True)

    @staticmethod
    def int_to_bytestring(i, padding=8):
        """
        Turns an integer to the OATH specified
        bytestring, which is fed to the HMAC
        along with the secret
        """
        result = bytearray()
        while i != 0:
            result.append(i & 0xFF)
            i >>= 8
        return bytes(bytearray(reversed(result)).rjust(padding, b'\0'))

    def generateTOTP(self):
        timecode = int(time.mktime(datetime.datetime.now().timetuple())/30)
        hasher = hmac.new(self.byte_secret(), self.int_to_bytestring(timecode), hashlib.sha1)
        hmac_hash = bytearray(hasher.digest())
        offset = hmac_hash[-1] & 0xf
        code = ((hmac_hash[offset] & 0x7f) << 24 |
                (hmac_hash[offset + 1] & 0xff) << 16 |
                (hmac_hash[offset + 2] & 0xff) << 8 |
                (hmac_hash[offset + 3] & 0xff))
        codelength = 6
        str_code = str(code % 10 ** codelength)
        while len(str_code) < codelength:
            str_code = '0' + str_code
        return str_code


#parses response from gateway

#function names are just lowercase types, so for type GUILD_MEMBER_LIST_UPDATE, the function is guild_member_list_update
class GuildParse(object):
	@staticmethod
	def guild_member_list_update(response):
		memberdata = {
		    "online_count": response["d"]["online_count"],
		    "member_count": response["d"]["member_count"],
		    "id": response["d"]["id"],
		    "guild_id": response["d"]["guild_id"],
		    "hoisted_roles": response["d"]["groups"],
		    "types": [],
		    "locations": [],
		    "updates": []
		}
		#now time to look over ops
		for chunk in response['d']['ops']:
			memberdata['types'].append(chunk['op'])
			if chunk['op'] in ('SYNC', 'INVALIDATE'):
				memberdata['locations'].append(chunk['range'])
				if chunk['op'] == 'SYNC':
					memberdata['updates'].append(chunk['items'])
				else: #invalidate
					memberdata['updates'].append([])
			elif chunk['op'] in ('INSERT', 'UPDATE', 'DELETE'): #only update the 0,99 range btw
				memberdata['locations'].append(chunk['index'])
				if chunk['op'] == 'DELETE':
					memberdata['updates'].append([])
				else:
					memberdata['updates'].append(chunk['item'])
		#and that's it. mostly some renaming and shuffling around to help with usage and reading :)
		return memberdata

	@staticmethod
	def guild_create(response, my_user_id):
		guilddata = dict(response["d"])
		#take care of position
		guilddata["my_data"] = response["members"]
		guilddata.pop("members") #this is actually not the member list, lol. But it usually contains our position/role in that guild so...still good info
		for index,member in enumerate(guilddata["my_data"]): 
			if member["user"]["id"] == my_user_id:
				guilddata["my_data"][index].pop("user")
				guilddata["my_data"][index]["user_id"] = my_user_id
			else:
				del guilddata["my_data"][index]
		#take care of emojis
		guilddata["emojis"] = {i["id"]:i for i in response["emojis"]}
		#take care of roles
		guilddata["roles"] = {j["id"]:j for j in response["roles"]}
		#take care of channels
		guilddata["channels"] = {k["id"]:k for k in response["channels"]}
		return guilddata


#parses response from gateway
from .start.parse import StartParse
from .guild.parse import GuildParse
from .user.parse import UserParse
from .messages.parse import MessageParse

import copy

#function names are just lowercase types, so for type GUILD_MEMBER_LIST_UPDATE, the function is guild_member_list_update
class Parse(object):
	def __init__(self, response):
		self.response = copy.deepcopy(response)

	def auto(self): #auto parse, does not allow for custom inputs
		if hasattr(self, self.response['t'].lower()):
			return getattr(self, self.response['t'].lower())()
		return self.response['d'] #just return the value of d if there's no parse function for it yet

	def ready(self):
		return StartParse.ready(self.response)

	def ready_supplemental(self):
		return StartParse.ready_supplemental(self.response)

	def guild_member_list_update(self):
		return GuildParse.guild_member_list_update(self.response)

	def guild_create(self, my_user_id="0"): #personal user id needed to update personal roles for that guild
		return GuildParse.guild_create(self.response, my_user_id)

	def message_create(self):
		return MessageParse.message_create(self.response)

	def sessions_replace(self, session_id="0"):
		return UserParse.sessions_replace(self.response, session_id)



from .start.login import Login
from .start.superproperties import SuperProperties
from .start.other import Other
from .start.totp import TOTP

from .guild.guild import Guild
from .messages.messages import Messages
from .messages.embed import Embedder
from .user.user import User
from .stickers.stickers import Stickers
from .science.science import Science

from .gateway.gateway import *

import time
import base64
import requests
import random

class Client:
    def __init__(self, email="", password="", secret="", code="", token="", proxy_host=None, proxy_port=None, user_agent="random", locale="en-US", build_num="request", log=True):
        #step 1: vars
        self.log = log
        self.locale = locale
        self.__user_token = token
        self.__user_email = email
        self.__user_password = password
        self.__totp_secret = secret
        self.__xfingerprint = ""
        self.userData = {}
        self.__proxy_host = None if proxy_host in (None,False) else proxy_host
        self.__proxy_port = None if proxy_port in (None,False) else proxy_port
        self.discord = 'https://discord.com/api/v8/'
        self.websocketurl = 'wss://gateway.discord.gg/?encoding=json&v=8&compress=zlib-stream'
        #step 2: user agent
        if user_agent != "random":
            self.__user_agent = user_agent
        else:
            import random_user_agent.user_agent #only really want to import this if needed
            self.__user_agent = random_user_agent.user_agent.UserAgent(limit=100).get_random_user_agent()
            if self.log: print('Randomly generated user agent: '+self.__user_agent)
        #step 3: http request headers
        headers = {
            "Origin": "https://discord.com",
            "User-Agent": self.__user_agent,
            "Accept": "*/*",
            "Accept-Encoding": "gzip, deflate, br",
            "Accept-Language": self.locale,
            "Cache-Control": "no-cache",
            "Pragma": "no-cache",
            "Referer": "https://discord.com/channels/@me",
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch_Site": "same-origin",
            "Connection": "keep-alive",
            "Content-Type": "application/json"
        }
        self.s = requests.Session()
        self.s.headers.update(headers)
        if self.__proxy_host != None: #self.s.proxies defaults to {}
            self.proxies = {
            'http': "http://" +  self.__proxy_host+':'+self.__proxy_port,
            'https': "https://" +  self.__proxy_host+':'+self.__proxy_port
            }
            self.s.proxies.update(self.proxies)
        #step 4: cookies
        self.s.cookies.update({"locale": self.locale})
        #step 5: super-properties (part of headers)
        self.__super_properties = SuperProperties(self.s, buildnum=build_num, log=self.log).GetSuperProperties(self.__user_agent)
        self.s.headers.update({"X-Super-Properties": base64.b64encode(str(self.__super_properties).encode()).decode("utf-8")})
        #step 6: token/authorization/fingerprint (also part of headers, except for fingerprint)
        if self.__user_token in ("",None,False): #assuming email and pass are given...
            self.__user_token, self.__xfingerprint = Login(self.s, self.discord, self.log).GetToken(email=email, password=password, secret=secret, code=code) #update token from "" to actual value
            time.sleep(1)            
        self.s.headers.update({"Authorization": self.__user_token}) #update headers
        #step 7: gateway (object initialization)
        self.gateway = GatewayServer(self.websocketurl, self.__user_token, self.__super_properties, self.s, self.discord, self.log) #self.s contains proxy host and proxy port already
        #step 8: embed (object initialization)
        self.Embedder = Embedder
        #step 9: somewhat prepare for science events
        self.Science = ""

##########################################################

    '''
    test connection
    '''
    def connectionTest(self):
        url=self.discord+'users/@me?with_analytics_token=true'
        connection = self.s.get(url)
        if connection.status_code == 200:
            if self.log: print("Connected")
            self.userData = connection.json()
        else:
            if self.log: print("Incorrect Token")
        return connection

    '''
    discord snowflake to unix timestamp and back
    '''
    def snowflake_to_unixts(self,snowflake):
        return (int(snowflake)/4194304+1420070400000)/1000

    def unixts_to_snowflake(self,unixts):
        return (int(unixts)*1000-1420070400000)*4194304

    '''
    start
    '''
    def login(self, email, password, undelete=False, captcha=None, source=None, gift_code_sku_id=None, secret="", code=""):
        return Login(self.s, self.discord, self.log).GetToken(email, password, undelete, captcha, source, gift_code_sku_id, secret, code)

    def getXFingerprint(self):
        return Login(self.s, self.discord, self.log).GetXFingerprint()

    def getBuildNumber(self):
        return SuperProperties(self.s, "request", self.log).RequestBuildNumber()

    def getSuperProperties(self, user_agent, buildnum="request"):
        return SuperProperties(self.s, buildnum, self.log).GetSuperProperties(user_agent)

    def getGatewayUrl(self):
        return Other(self.s, self.discord, self.log).getGatewayUrl()

    def getDiscordStatus(self):
        return Other(self.s, self.discord, self.log).getDiscordStatus()

    def getDetectables(self):
        return Other(self.s, self.discord, self.log).getDetectables()

    def getOauth2Tokens(self):
        return Other(self.s, self.discord, self.log).getOauth2Tokens()

    def getVersionStableHash(self, underscore=None):
        return Other(self.s, self.discord, self.log).getVersionStableHash(underscore)

    '''
    Messages
    '''
    #create DM
    def createDM(self,recipients):
        return Messages(self.discord,self.s,self.log).createDM(recipients)

    #get recent messages
    def getMessages(self,channelID,num=1,beforeDate=None,aroundMessage=None): # num <= 100, beforeDate is a snowflake
        return Messages(self.discord,self.s,self.log).getMessages(channelID,num,beforeDate,aroundMessage)

    #get message by channel ID and message ID
    def getMessage(self, channelID, messageID):
        return Messages(self.discord,self.s,self.log).getMessage(channelID, messageID)

    #send messages
    def sendMessage(self, channelID, message, nonce="calculate", tts=False, embed=None, message_reference=None, allowed_mentions=None, sticker_ids=None):
        return Messages(self.discord,self.s,self.log).sendMessage(channelID, message, nonce, tts, embed, message_reference, allowed_mentions, sticker_ids)

    #send files (local or link)
    def sendFile(self,channelID,filelocation,isurl=False,message="", tts=False, message_reference=None, sticker_ids=None):
        return Messages(self.discord,self.s,self.log).sendFile(channelID,filelocation,isurl,message, tts, message_reference, sticker_ids)

    #reply, with a message and/or file
    def reply(self, channelID, messageID, message, nonce="calculate", tts=False, embed=None, allowed_mentions={"parse":["users","roles","everyone"],"replied_user":False}, sticker_ids=None, file=None, isurl=False):
        return Messages(self.discord,self.s,self.log).reply(channelID, messageID, message, nonce, tts, embed, allowed_mentions, sticker_ids, file, isurl)

    #search messages
    def searchMessages(self,guildID,channelID=[],userID=[],mentionsUserID=[],has=[],beforeDate=None,afterDate=None,textSearch=None,afterNumResults=None, limit=None, extraParams=""):
        return Messages(self.discord,self.s,self.log).searchMessages(guildID,channelID,userID,mentionsUserID,has,beforeDate,afterDate,textSearch,afterNumResults,limit,extraParams)

    #filter searchMessages, takes in the output of searchMessages (a requests response object) and outputs a list of target messages
    def filterSearchResults(self,searchResponse):
        return Messages(self.discord,self.s,self.log).filterSearchResults(searchResponse)

    #sends the typing action for 10 seconds (or technically until you change the page)
    def typingAction(self,channelID):
        return Messages(self.discord,self.s,self.log).typingAction(channelID)

    #delete message
    def deleteMessage(self,channelID,messageID):
        return Messages(self.discord,self.s,self.log).deleteMessage(channelID,messageID)

    #edit message
    def editMessage(self,channelID,messageID,newMessage):
        return Messages(self.discord,self.s,self.log).editMessage(channelID, messageID, newMessage)

    #pin message
    def pinMessage(self,channelID,messageID):
        return Messages(self.discord,self.s,self.log).pinMessage(channelID,messageID)

    #un-pin message
    def unPinMessage(self,channelID,messageID):
        return Messages(self.discord,self.s,self.log).unPinMessage(channelID,messageID)

    #get pinned messages
    def getPins(self,channelID):
        return Messages(self.discord,self.s,self.log).getPins(channelID)

    #add reaction
    def addReaction(self,channelID,messageID,emoji):
        return Messages(self.discord,self.s,self.log).addReaction(channelID,messageID,emoji)

    #remove reaction
    def removeReaction(self,channelID,messageID,emoji):
        return Messages(self.discord,self.s,self.log).removeReaction(channelID,messageID,emoji)

    #acknowledge message (mark message read)
    def ackMessage(self,channelID,messageID,ackToken=None):
        return Messages(self.discord,self.s,self.log).ackMessage(channelID,messageID,ackToken)

    #unacknowledge message (mark message unread)
    def unAckMessage(self,channelID,messageID,numMentions=0):
        return Messages(self.discord,self.s,self.log).unAckMessage(channelID,messageID,numMentions)

    def bulkAck(self, data):
        return Messages(self.discord,self.s,self.log).bulkAck(data)

    def getTrendingGifs(self, provider="tenor", locale="en-US", media_format="mp4"):
        return Messages(self.discord,self.s,self.log).getTrendingGifs(provider, locale, media_format)

    '''
    Stickers
    '''
    def getStickers(self, directoryID="758482250722574376", store_listings=False, locale="en-US"):
        return Stickers(self.discord,self.s,self.log).getStickers(directoryID, store_listings, locale)

    def getStickerFile(self, stickerID, stickerAsset): #this is an animated png
        return Stickers(self.discord,self.s,self.log).getStickerFile(stickerID, stickerAsset)

    def getStickerJson(self, stickerID, stickerAsset):
        return Stickers(self.discord,self.s,self.log).getStickerJson(stickerID, stickerAsset)

    def getStickerPack(self, stickerPackID):
        return Stickers(self.discord,self.s,self.log).getStickerPack(stickerPackID)

    '''
    User relationships
    '''
    #create outgoing friend request
    def requestFriend(self,user): #you can input a userID(snowflake) or a user discriminator
        return User(self.discord,self.s,self.log).requestFriend(user)

    #accept incoming friend request
    def acceptFriend(self,userID):
        return User(self.discord,self.s,self.log).acceptFriend(userID)

    #remove friend OR unblock user
    def removeRelationship(self,userID):
        return User(self.discord,self.s,self.log).removeRelationship(userID)

    #block user
    def blockUser(self,userID):
        return User(self.discord,self.s,self.log).blockUser(userID)

    '''
    Profile edits
    '''
    # set status; moved to gateway; dont run this function by itself
    def setStatusHelper(self,status):
        return User(self.discord,self.s,self.log).setStatusHelper(status)

    # set avatar
    def setAvatar(self,imagePath):
        return User(self.discord,self.s,self.log).setAvatar(imagePath)

    #set username
    def setUsername(self, username): #USER PASSWORD NEEDS TO BE SET BEFORE THIS IS RUN
        return User(self.discord,self.s,self.log).setUsername(username, password=self.__user_password)

    #set email
    def setEmail(self, email): #USER PASSWORD NEEDS TO BE SET BEFORE THIS IS RUN
        return User(self.discord,self.s,self.log).setEmail(email, password=self.__user_password)

    #set password
    def setPassword(self, new_password): #USER PASSWORD NEEDS TO BE SET BEFORE THIS IS RUN
        return User(self.discord,self.s,self.log).setPassword(new_password, password=self.__user_password)

    #set discriminator
    def setDiscriminator(self, discriminator): #USER PASSWORD NEEDS TO BE SET BEFORE THIS IS RUN
        return User(self.discord,self.s,self.log).setDiscriminator(discriminator, password=self.__user_password)

    '''
    other user stuff
    '''
    def getProfile(self, userID):
        return User(self.discord,self.s,self.log).getProfile(userID)

    def info(self, with_analytics_token=None):
        return User(self.discord,self.s,self.log).info(with_analytics_token)

    def getUserAffinities(self):
        return User(self.discord,self.s,self.log).getUserAffinities()

    def getGuildAffinities(self):
        return User(self.discord,self.s,self.log).getGuildAffinities()

    def getMentions(self, limit=25, roleMentions=True, everyoneMentions=True):
        return User(self.discord,self.s,self.log).getMentions(limit, roleMentions, everyoneMentions)

    def removeMentionFromInbox(self, messageID):
        return User(self.discord,self.s,self.log).removeMentionFromInbox(messageID)

    def setHypesquad(self, house):
        return User(self.discord,self.s,self.log).setHypesquad(house)

    def leaveHypesquad(self):
        return User(self.discord,self.s,self.log).leaveHypesquad()

    def setLocale(self, locale):
        response = User(self.discord,self.s,self.log).setLocale(locale)
        self.locale = locale
        self.s.headers["Accept-Language"] = self.locale
        self.s.cookies["locale"] = self.locale

    def calculateTOTPcode(self, secret="default"): #need to put this function here (instead of in login folder or user folder) because it updates the secret (if and only if secret == "")
        if secret == "default":
            if self.__totp_secret == "":
                self.__totp_secret = ''.join(random.choice(list('ABCDEFGHIJKLMNOPQRSTUVWXYZ234567')) for _ in range(16)) #random base32 (len 16)
            secret = self.__totp_secret
        if "?secret=" in secret:
            secret = secret[secret.index("?secret=")+8: secret.index("?secret=")+24]
        return TOTP(secret).generateTOTP(), secret #secret is returned just in case it wasn't set at the beginning.

    def getTOTPurl(self, secret): #use this to store your totp secret/qr pic; btw url format is otpauth://totp/Discord:EMAIL?secret=SECRET&issuer=Discord
        url = "otpauth://totp/Discord"
        if self.__user_email != "":
            url += ":"+self.__user_email
        url += "?secret="+secret+"&issuer=Discord"
        return url

    def enable2FA(self): #this also returns backup codes (value of key "backup_codes"). USER PASSWORD NEEDS TO BE SET BEFORE THIS IS RUN
        code = self.calculateTOTPcode()[0]
        result = User(self.discord,self.s,self.log).enable2FA(code, secret=self.__totp_secret, password=self.__user_password)
        self.__user_token = result.json()["token"]
        self.s.headers['Authorization'] = self.__user_token
        return result

    def disable2FA(self, code="calculate", clearSecretAfter=False): #either set your token before running this or input a code.
        if code == "calculate":
            code = self.calculateTOTPcode()[0] #this will generate a random secret if you dont have one set, so...set your secret before running this
        code = str(code) #just in case
        result = User(self.discord,self.s,self.log).disable2FA(code)
        self.__user_token = result.json()["token"]
        self.s.headers['Authorization'] = self.__user_token
        if clearSecretAfter: #this is dangerous (even though disable2FA should error out before getting to this point if something goes wrong). If you already have your secret saved, clear away. By default this is set to False to avoid any mishaps.
            self.__totp_secret = ""
        return result

    def getRTCregions(self):
        return User(self.discord,self.s,self.log).getRTCregions()

    def setAFKtimeout(self, timeout_seconds):
        return User(self.discord,self.s,self.log).setAFKtimeout(timeout_seconds)

    def setTheme(self, theme): #"light" or "dark"
        return User(self.discord,self.s,self.log).setTheme(theme)

    def setMessageDisplay(self, CozyOrCompact): #"cozy" or "compact"
        return User(self.discord,self.s,self.log).setMessageDisplay(CozyOrCompact)

    def enableDevMode(self, enable): #boolean, default=False
        return User(self.discord,self.s,self.log).enableDevMode(enable)

    def activateApplicationTestMode(self, applicationID):
        return User(self.discord,self.s,self.log).activateApplicationTestMode(applicationID)

    def getApplicationData(self, applicationID, with_guild=False):
        return User(self.discord,self.s,self.log).getApplicationData(applicationID, with_guild)

    def getBackupCodes(self, regenerate=False):
        return User(self.discord,self.s,self.log).getBackupCodes(self.__user_password, regenerate)

    def enableInlineMedia(self, enable): #boolean, default=True
        return User(self.discord,self.s,self.log).enableInlineMedia(enable)

    def enableLargeImagePreview(self, enable): #boolean, default=True
        return User(self.discord,self.s,self.log).enableLargeImagePreview(enable)

    def enableGifAutoPlay(self, enable): #boolean, default=True
        return User(self.discord,self.s,self.log).enableGifAutoPlay(enable)

    def enableLinkPreview(self, enable): #boolean, default=True
        return User(self.discord,self.s,self.log).enableLinkPreview(enable)

    def enableReactionRendering(self, enable): #boolean, default=True
        return User(self.discord,self.s,self.log).enableReactionRendering(enable)

    def enableAnimatedEmoji(self, enable): #boolean, default=True
        return User(self.discord,self.s,self.log).enableAnimatedEmoji(enable)

    def enableEmoticonConversion(self, enable): #boolean, default=True
        return User(self.discord,self.s,self.log).enableEmoticonConversion(enable)

    def setStickerAnimation(self, setting): #string, default="always"
        return User(self.discord,self.s,self.log).setStickerAnimation(setting)

    def enableTTS(self, enable): #boolean, default=True
        return User(self.discord,self.s,self.log).enableTTS(enable)

    def getBillingHistory(self, limit=20):
        return User(self.discord,self.s,self.log).getBillingHistory(limit)

    def getPaymentSources(self):
        return User(self.discord,self.s,self.log).getPaymentSources()

    def getBillingSubscriptions(self):
        return User(self.discord,self.s,self.log).getBillingSubscriptions()

    def getStripeClientSecret(self):
        return User(self.discord,self.s,self.log).getStripeClientSecret()

    def logout(self, provider=None, voip_provider=None):
        return User(self.discord,self.s,self.log).logout(provider, voip_provider)

    '''
    Guild/Server stuff
    '''
    #get guild info from invite code
    def getInfoFromInviteCode(self,inviteCode):
        return Guild(self.discord,self.s,self.log).getInfoFromInviteCode(inviteCode)

    #join guild with invite code
    def joinGuild(self,inviteCode):
        return Guild(self.discord,self.s,self.log).joinGuild(inviteCode)

    #leave guild
    def leaveGuild(self, guildID):
        return Guild(self.discord,self.s,self.log).leaveGuild(guildID)

    #create invite
    def createInvite(self, channelID, max_age_seconds=False, max_uses=False, grantTempMembership=False, checkInvite="", targetType=""):
        return Guild(self.discord,self.s,self.log).createInvite(channelID, max_age_seconds, max_uses, grantTempMembership, checkInvite, targetType)

    #kick a user
    def kick(self,guildID,userID,reason=""):
        return Guild(self.discord,self.s,self.log).kick(guildID,userID,reason)

    #ban a user
    def ban(self,guildID,userID,deleteMessagesDays=0,reason=""):
        return Guild(self.discord,self.s,self.log).ban(guildID,userID,deleteMessagesDays,reason)

    #unban a user
    def revokeBan(self, guildID, userID):
        return Guild(self.discord,self.s,self.log).revokeBan(guildID, userID)

    #look up a user in a guild
    def getGuildMember(self,guildID,userID):
        return Guild(self.discord,self.s,self.log).getGuildMember(guildID,userID)

    def getMemberVerificationData(self, guildID, with_guild=False, invite_code=None):
        return Guild(self.discord,self.s,self.log).getMemberVerificationData(guildID, with_guild, invite_code)

    def agreeGuildRules(self, guildID, form_fields, version="2021-01-05T01:44:32.163000+00:00"):
        return Guild(self.discord,self.s,self.log).agreeGuildRules(guildID, form_fields, version)

    '''
    "Science", aka Discord's tracking endpoint (https://luna.gitlab.io/discord-unofficial-docs/science.html - "Discord argues that they need to collect the data in the case the User allows the usage of the data later on. Which in [luna's] opinion is complete bullshit. Have a good day.")
    '''
    def science(self, events): #the real prep for science events happens down here, and only once for each client obj
        if self.Science == "":
            try:
                #get xfingerprint
                if self.__xfingerprint == "":
                    self.__xfingerprint = Login(self.s, self.discord, self.log).GetXFingerprint()
                    time.sleep(1)
                #get analytics token
                self.userData = User(self.discord,self.s,self.log).info(with_analytics_token=True).json() #this is essentially the connection test. We need it cause we can get important data without connecting to the gateway.
                connectiontest = self.userData["analytics_token"]
            except:
                self.userData = {"analytics_token": None, "id": "0"} #if token invalid
            #initialize Science object
            self.Science = Science(self.discord, self.s, self.log, self.userData["analytics_token"], self.userData["id"], self.__xfingerprint)
        return self.Science.science(events)

    def calculateClientUUID(self, eventNum="default", userID="default", increment=True):
        return self.Science.UUIDobj.calculate(eventNum, userID, increment)

    def refreshClientUUID(self, resetEventNum=True):
        return self.Science.UUIDobj.refresh(resetEventNum)

    def parseClientUUID(self, client_uuid):
        return self.Science.UUIDobj.parse(client_uuid)



#combo
import datetime #eh will need this later

class UserCombo(object):
	def __init__(self, gatewayobj):
		self.gatewayobj = gatewayobj

	def constructActivitiesList(self, updates={}, remove=None): #update is a bit too raw, will fix later
		currentActivities = self.gatewayobj.session.userSettings["activities"]		
		activities = {4: {}, 0: {}, 1: {}, 2: {}, 3: {}}
		for i in currentActivities:
			if i==4 and remove!=i:
				activities[4]["type"] = 4
				activities[4]["state"] = currentActivities[4]["state"]
				activities[4]["name"] = "Custom Status"
				activities[4]["emoji"] = None #idk will fix later
			elif remove!=i: #idk will fix later
				activities[i]["type"] = i
				activities[i]["name"] = currentActivities[i]["name"]
				activities[i]["timestamps"] = currentActivities[i]["timestamps"]
		activities.update(updates)
		activitiesList = []
		for j in activities:
			if len(activities[j])>0:
				activitiesList.append(activities[j])
		return activitiesList

	def setStatus(self, status):
		if status != self.gatewayobj.session.userSettings["status"]:
			self.gatewayobj.request.setStatus(status, self.constructActivitiesList())

	#def setPlayingStatus(self, game):

	def removePlayingStatus(self):
		currentStatus = self.gatewayobj.session.userSettings["status"]
		self.gatewayobj.request.setStatus(currentStatus, self.constructActivitiesList(remove=0))

	#def setStreamingStatus(self, stream, metadata={}): 

	def removeStreamingStatus(self):
		currentStatus = self.gatewayobj.session.userSettings["status"]
		self.gatewayobj.request.setStatus(currentStatus, self.constructActivitiesList(remove=1))

	#def setListeningStatus(self, song, metadata={}): 

	def removeListeningStatus(self):
		currentStatus = self.gatewayobj.session.userSettings["status"]
		self.gatewayobj.request.setStatus(currentStatus, self.constructActivitiesList(remove=2))

	#def setWatchingStatus(self, show, metadata={}): 

	def removeWatchingStatus(self):
		currentStatus = self.gatewayobj.session.userSettings["status"]
		self.gatewayobj.request.setStatus(currentStatus, self.constructActivitiesList(remove=3))

	def setCustomStatus(self, customstatus):
		currentStatus = self.gatewayobj.session.userSettings["status"]
		self.gatewayobj.request.setStatus(currentStatus, self.constructActivitiesList(updates={4:{"type":4, "state": customstatus, "name":"Custom Status", "emoji": None}}))

	def removeCustomStatus(self):
		currentStatus = self.gatewayobj.session.userSettings["status"]
		self.gatewayobj.request.setStatus(currentStatus, self.constructActivitiesList(remove=4))


#parses response from gateway

class UserParse(object):
	@staticmethod
	def sessions_replace(response, session_id):
		importantdata = {}
		activeCounter = {} #priority = 0
		allCounter = {} #priority = 1
		sessionidCounter = {} #priority = 2
		#sessions_replace is one of those undocumented events that have weird formatting. :(
		for session in response['d']:
			if session.get("active") == True:
				activeCounter = session
				break; #no need to check anything else
			elif session.get("session_id") == "all":
				allCounter = session
			elif session.get("session_id") == session_id:
				sessionidCounter = session
		#now start the processing
		if len(activeCounter) > 0:
			importantdata["status"] = activeCounter["status"]
			importantdata["activities"] = {i["type"]:i for i in activeCounter["activities"]}
			return importantdata
		elif len(allCounter) > 0:
			importantdata["status"] = allCounter["status"]
			importantdata["activities"] = {j["type"]:j for j in allCounter["activities"]}
			return importantdata
		elif len(sessionidCounter) > 0:
			importantdata["status"] = sessionidCounter["status"]
			importantdata["activities"] = {k["type"]:k for k in sessionidCounter["activities"]}
			return importantdata
		else:
			return {}



class Session:
    settings_ready = {}
    settings_ready_supp = {}
    def __init__(self, input_settings_ready, input_settings_ready_supp):
        Session.settings_ready = input_settings_ready
        Session.settings_ready_supp = input_settings_ready_supp
        self.guild = guild
        self.DM = DM
        self.relationship = relationship
        self.userGuildSetting = userGuildSetting

    def read(self): #returns all Session settings
        return [self.settings_ready, self.settings_ready_supp]

    ###***USER***###
    @property
    def user(self):
        return self.settings_ready['user']

    ###***GUILDS***### (general)
    @property
    def guilds(self):
        return self.settings_ready['guilds']

    @property
    def allGuildIDs(self): #even if you're not in that guild
        return list(self.settings_ready['guilds'].keys())

    @property
    def guildIDs(self): #only for guilds that you're in
        return [guildID for guildID in self.settings_ready['guilds'].keys() if "removed" not in self.settings_ready['guilds'][guildID]]

    def setGuildData(self, guildID, guildData):
        self.settings_ready['guilds'][guildID] = guildData

    ###***RELATIONSHIPS***### (general)
    @property
    def relationships(self):
        return self.settings_ready['relationships']
    
    @property
    def relationshipIDs(self):
        return list(self.settings_ready['relationships'].keys())

    #friends
    @property
    def friends(self):
        f = {}
        for i in self.relationships: #where i is a user id
            if self.relationships[i]['type'] == 1:
                f[i] = self.relationships[i]
        return f
    
    @property
    def friendIDs(self):
        return list(self.friends.keys())

    #blocked    
    @property
    def blocked(self):
        b = {}
        for i in self.relationships: #where i is a user id
            if self.relationships[i]['type'] == 2:
                b[i] = self.relationships[i]
        return b
    
    @property
    def blockedIDs(self):
        return list(self.blocked.keys())
    
    #incoming    
    @property
    def incomingFriendRequests(self):
        ifr = {}
        for i in self.relationships:
            if self.relationships[i]['type'] == 3:
                ifr[i] = self.relationships[i]
        return ifr
    
    @property
    def incomingFriendRequestIDs(self):
        return list(self.incomingFriendRequests.keys())

    #outgoing
    @property
    def outgoingFriendRequests(self):
        ofr = {}
        for i in self.relationships:
            if self.relationships[i]['type'] == 4:
                ofr[i] = self.relationships[i]
        return ofr
    
    @property
    def outgoingFriendRequestIDs(self):
        return list(self.outgoingFriendRequests.keys())

    #friend merged presences    
    @property
    def onlineFriends(self):
        return self.settings_ready_supp['online_friends']
    
    @property
    def onlineFriendIDs(self):
        return list(self.onlineFriends.keys())
        

    ###***DMs***### (general)
    @property
    def DMs(self):
        return self.settings_ready['private_channels']

    @property
    def DMIDs(self):
        return list(self.DMs.keys())
        

    ###***USER SETTINGS***### (general)
    @property
    def userGuildSettings(self):
        return self.settings_ready['user_guild_settings'] #so uh...this is not only for guilds. It also covers group DMs so uh yea...weird naming
    
    @property
    def userSettings(self):
        return self.settings_ready['user_settings']
    
    @property
    def optionsForUserSettings(self):
        return list(self.settings_ready['user_settings'].keys())
        
    def updateUserSettings(self, data):
        self.settings_ready['user_settings'].update(data)

    ###other stuff
    @property
    def analyticsToken(self):
        return self.settings_ready['analytics_token']

    @property
    def connectedAccounts(self):
        return self.settings_ready['connected_accounts']

    @property
    def consents(self):
        return self.settings_ready['consents']

    @property
    def experiments(self):
        return self.settings_ready['experiments']

    @property
    def friendSuggestionCount(self):
        return self.settings_ready['friend_suggestion_count']

    @property
    def guildExperiments(self):
        return self.settings_ready['guild_experiments']

    @property
    def readStates(self):
        return self.settings_ready['read_state'] #another advantage of using websockets instead of requests (see https://github.com/discord/discord-api-docs/issues/13)

    @property
    def geoOrderedRtcRegions(self):
        return self.settings_ready['geo_ordered_rtc_regions']

    @property
    def cachedUsers(self): #idk what these are
        return self.settings_ready['users']

    @property
    def tutorial(self):
        return self.settings_ready['tutorial'] #that tutorial you get when you first make an account


###specific guild
class guild(Session):
    def __init__(self, guildID):
        self.guildID = guildID

    @property
    def data(self): #self.settings_ready['guilds']
        return Session.settings_ready['guilds'][self.guildID]

    def setData(self, newData):
        Session.settings_ready['guilds'][self.guildID] = newData

    def updateData(self, data):
        Session.settings_ready['guilds'][self.guildID].update(data)

    @property
    def unavailable(self):
        return 'unavailable' in Session.settings_ready['guilds'][self.guildID]

    @property
    def hasMembers(self):
        return len(Session.settings_ready['guilds'][self.guildID]['members']) == 0

    @property
    def members(self):
        return Session.settings_ready['guilds'][self.guildID]['members']

    @property
    def memberIDs(self):
        return list(self.members.keys())

    def resetMembers(self):
        Session.settings_ready['guilds'][self.guildID]['members'] = {}

    def updateOneMember(self, userID, userProperties):
        Session.settings_ready['guilds'][self.guildID]['members'][userID] = userProperties

    def updateMembers(self, memberdata): #where member data is a dictionary --> {userId: {properties}, ...}
        Session.settings_ready['guilds'][self.guildID]['members'].update(memberdata)

    @property
    def owner(self):
        return Session.settings_ready['guilds'][self.guildID]['owner_id'] #returns type int

    @property
    def boostLvl(self):
        return Session.settings_ready['guilds'][self.guildID]['premium_tier']

    @property
    def emojis(self):
        return Session.settings_ready['guilds'][self.guildID]['emojis']

    @property
    def emojiIDs(self):
        return list(self.emojis.keys())

    @property
    def banner(self):
        return Session.settings_ready['guilds'][self.guildID]['banner'] #if returns 'nil' then there's no banner

    @property
    def discoverySplash(self): #not sure what this is about, something about server discoverability i guess (https://discord.com/developers/docs/resources/guild)
        return Session.settings_ready['guilds'][self.guildID]['discovery_splash']

    @property
    def msgNotificationSettings(self): #returns an int, 0=all messages, 1=only mentions (https://discord.com/developers/docs/resources/guild#guild-object-default-message-notification-level)
        return Session.settings_ready['guilds'][self.guildID]['default_message_notifications']

    @property
    def rulesChannelID(self): #nil if no rules channel id, idk if this always works so it might actually be more useful just to look for the word "rules" in channel names
        return Session.settings_ready['guilds'][self.guildID]['rules_channel_id']

    @property
    def verificationLvl(self): #returns an int, 0-4 (https://discord.com/developers/docs/resources/guild#guild-object-verification-level)
        return Session.settings_ready['guilds'][self.guildID]['verification_level']

    @property
    def features(self): #returns a list of strings (https://discord.com/developers/docs/resources/guild#guild-object-guild-features)
        return Session.settings_ready['guilds'][self.guildID]['features']

    @property
    def joinTime(self): #returns when you joined the server, type string
        return Session.settings_ready['guilds'][self.guildID]['joined_at']

    @property
    def region(self):
        return Session.settings_ready['guilds'][self.guildID]['region']

    @property
    def applicationID(self): #returns application id of the guild creator if it is bot-created (https://discord.com/developers/docs/resources/guild#guild-object-guild-features)
        return Session.settings_ready['guilds'][self.guildID]['application_id']

    @property
    def afkChannelID(self): #not sure what this is
        return Session.settings_ready['guilds'][self.guildID]['afk_channel_id']

    @property
    def icon(self): #https://discord.com/developers/docs/reference#image-formatting
        return Session.settings_ready['guilds'][self.guildID]['icon']

    @property
    def name(self):
        return Session.settings_ready['guilds'][self.guildID]['name']

    @property
    def maxVideoChannelUsers(self):
        return Session.settings_ready['guilds'][self.guildID]['max_video_channel_users']

    @property
    def roles(self): #https://discord.com/developers/docs/topics/permissions#role-object
        return Session.settings_ready['guilds'][self.guildID]['roles']

    @property
    def publicUpdatesChannelID(self):
        return Session.settings_ready['guilds'][self.guildID]['public_updates_channel_id']

    @property
    def systemChannelFlags(self): #https://discord.com/developers/docs/resources/guild#guild-object-system-channel-flags
        return Session.settings_ready['guilds'][self.guildID]['system_channel_flags']

    @property
    def mfaLvl(self): #https://discord.com/developers/docs/resources/guild#guild-object-mfa-level
        return Session.settings_ready['guilds'][self.guildID]['mfa_level']

    @property
    def afkTimeout(self): #returns type int, unit seconds, https://discord.com/developers/docs/resources/guild
        return Session.settings_ready['guilds'][self.guildID]['afk_timeout']

    @property
    def hashes(self): #https://github.com/discord/discord-api-docs/issues/1642
        return Session.settings_ready['guilds'][self.guildID]['guild_hashes']

    @property
    def systemChannelID(self): #returns an int, the id of the channel where guild notices such as welcome messages and boost events are posted
        return Session.settings_ready['guilds'][self.guildID]['system_channel_id']

    @property
    def lazy(self): #slightly different naming format since it returns a boolean (https://luna.gitlab.io/discord-unofficial-docs/lazy_guilds.html)
        return Session.settings_ready['guilds'][self.guildID]['lazy']

    @property
    def numBoosts(self): #get number of boosts the server has gotten
        return Session.settings_ready['guilds'][self.guildID]['premium_subscription_count']

    @property
    def large(self): #slightly different naming format since it returns a boolean, large if more than 250 members
        return Session.settings_ready['guilds'][self.guildID]['large']

    @property
    def explicitContentFilter(self): #https://discord.com/developers/docs/resources/guild#guild-object-explicit-content-filter-level
        return Session.settings_ready['guilds'][self.guildID]['explicit_content_filter']

    @property
    def splashHash(self): #not sure what this is for
        return Session.settings_ready['guilds'][self.guildID]['splash']

    @property
    def memberCount(self):
        return Session.settings_ready['guilds'][self.guildID]['member_count']

    @property
    def description(self):
        return Session.settings_ready['guilds'][self.guildID]['description']

    @property
    def vanityUrlCode(self):
        return Session.settings_ready['guilds'][self.guildID]['vanity_url_code']

    @property
    def preferredLocale(self):
        return Session.settings_ready['guilds'][self.guildID]['preferred_locale']

    @property
    def channelsAndCategories(self): #returns all categories and all channels, all the data about that, wall of data so it can be a bit overwhelming, useful if you want to check how many channels your server has since discord counts categories as channels
        return Session.settings_ready['guilds'][self.guildID]['channels']

    @property
    def channelAndCategoryIDs(self):
        return list(self.channelsAndCategories.keys())
    
    @property
    def categories(self): #all data about all guild categories, can be overwhelming
        all_categories = {}
        for i in self.channelsAndCategories: #https://discord.com/developers/docs/resources/channel#channel-object-channel-types
            if self.channelsAndCategories[i]['type'] == 4:
                all_categories[i] = self.channelsAndCategories[i]
        return all_categories

    @property
    def categoryIDs(self):
        return list(self.categories.keys())

    def categoryData(self, categoryID):
        return self.channelsAndCategories[categoryID]

    @property
    def channels(self): #all data about all guild channels, can be overwhelming
        all_channels = {}
        for i in self.channelsAndCategories: #https://discord.com/developers/docs/resources/channel#channel-object-channel-types
            if self.channelsAndCategories[i]['type'] != 4:
                all_channels[i] = self.channelsAndCategories[i]
        return all_channels

    @property
    def channelIDs(self):
        return list(self.channels.keys())

    def channelData(self, channelID):
        return self.channelsAndCategories[channelID]

    @property
    def voiceStates(self): #https://discord.com/developers/docs/resources/voice#voice-state-object
        return Session.settings_ready_supp['voice_states'][self.guildID]

    @property
    def position(self): #my roles, nick, etc in a guild
        return Session.settings_ready['guilds'][self.guildID]['my_data']

###specific relationship
class relationship(Session): #not the same organization as class guild
    def __init__(self, userID):
        self.userID = userID

    @property
    def data(self): #usernames and discriminators are no longer provided in this data
        return Session.settings_ready['relationships'][self.userID]

###specific DM
class DM(Session):
    def __init__(self, DMID):
        self.DMID = DMID

    @property
    def data(self):
        return Session.settings_ready['private_channels'][self.DMID]

    @property
    def recipients(self): #returns everyone in that DM except you
        return self.data['recipient_ids']

###specific User Guild Settings; keep in mind that user guild settings also includes some group dm notification settings stuff
class userGuildSetting(Session):
    def __init__(self, guildID):
        self.guildID = guildID

    @property
    def data(self):
        if len(Session.settings_ready['user_guild_settings']['entries']) == 0:
            return None
        for i in range(len(Session.settings_ready['user_guild_settings']['entries'])):
            if Session.settings_ready['user_guild_settings']['entries'][i]['guild_id'] == self.guildID:
                return Session.settings_ready['user_guild_settings']['entries'][i]
        return None



class MessageParse(object):
	@staticmethod
	def message_create(response):
		message = dict(response["d"])
		types = {
		    0: "default",
		    1: "recipient_added",
		    2: "recipient_removed",
		    3: "call",
		    4: "channel_name_changed",
		    5: "channel_icon_changed",
		    6: "channel_message_pinned",
		    7: "guild_member_joined",
		    8: "user_premium_guild_subscription",
		    9: "user_premium_guild_subscription_tier_1",
		    10: "user_premium_guild_subscription_tier_2",
		    11: "user_premium_guild_subscription_tier_3",
		    12: "channel_follow_added",
		    14: "guild_discovery_disqualified",
		    15: "guild_discovery_requalified",
		    19: "reply",
		    20: "application_command"
		}
		message["type"] = types[response["d"]["type"]] #number to str
		return message



from ..fileparse.fileparse import Fileparse
from requests_toolbelt import MultipartEncoder
import random,string
import os.path
import time, datetime
import json
from ..RESTapiwrap import *

if __import__('sys').version.split(' ')[0] < '3.0.0':
    from urllib import quote_plus, urlencode
    from urlparse import urlparse
else:
    from urllib.parse import quote_plus, urlparse, urlencode

class Messages(object):
    def __init__(self, discord, s, log): #s is the requests session object
        self.discord = discord
        self.s = s
        self.log = log

    def calculateNonce(self, date="now"):
        if date == "now":
            date = datetime.datetime.now()
        unixts = time.mktime(date.timetuple())
        return str((int(unixts)*1000-1420070400000)*4194304)

    #add DM
    def createDM(self,recipients):
        url = self.discord+"users/@me/channels"
        body = {"recipients": recipients}
        return Wrapper.sendRequest(self.s, 'post', url, body, log=self.log)

    #get messages
    def getMessages(self,channelID,num,beforeDate,aroundMessage): # num is between 1 and 100, beforeDate is a snowflake
        url = self.discord+"channels/"+channelID+"/messages?limit="+str(num)
        if beforeDate != None:
            url += "&before="+str(beforeDate)
        elif aroundMessage != None:
            url += "&around="+str(aroundMessage)
        return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

    #get message by channel ID and message ID
    def getMessage(self, channelID, messageID):
        url = self.discord+"channels/"+channelID+"/messages?limit=1&around="+messageID
        return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

    #text message
    def sendMessage(self, channelID, message, nonce, tts=False, embed=None, message_reference=None, allowed_mentions=None, sticker_ids=None):
        url = self.discord+"channels/"+channelID+"/messages"
        if nonce == "calculate":
            body = {"content": message, "tts": tts, "nonce": self.calculateNonce()}
        else:
            body = {"content": message, "tts": tts, "nonce": str(nonce)}
        if embed != None:
            body["embed"] = embed
        if message_reference != None:
            body["message_reference"] = message_reference
        if allowed_mentions != None:
            body["allowed_mentions"] = allowed_mentions
        if sticker_ids != None:
            body["sticker_ids"] = sticker_ids
        return Wrapper.sendRequest(self.s, 'post', url, body, log=self.log)

    #send file
    def sendFile(self,channelID,filelocation,isurl=False,message="", tts=False, message_reference=None, sticker_ids=None):
        mimetype, extensiontype, fd = Fileparse(self.s,self.log).parse(filelocation,isurl) #guess extension from file data
        if mimetype == 'invalid': #error out
            return
        if isurl: #get filename
            a = urlparse(filelocation)
            if len(os.path.basename(a.path))>0: #if everything is normal...
                filename = os.path.basename(a.path)
            else: 
                if mimetype == 'unsupported': #if filetype not detected and extension not given
                    filename = 'unknown'
                else: #if filetype detected but extension not given
                    filename = 'unknown.'+extensiontype
        else: #local file
            filename = os.path.basename(os.path.normpath(filelocation))
        #now time to post the file
        url = self.discord+'channels/'+channelID+'/messages'
        if isurl:
            payload = {"content":message,"tts":tts}
            if message_reference != None:
                payload["message_reference"] = message_reference
                payload["type"] = 19
            if sticker_ids != None:
                payload["sticker_ids"] = sticker_ids
            fields={"file":(filename,fd,mimetype), "payload_json":(None,json.dumps(payload))}
        else:
            payload = {"content":message,"tts":tts}
            if message_reference != None:
                payload["message_reference"] = message_reference
                payload["type"] = 19
            if sticker_ids != None:
                payload["sticker_ids"] = sticker_ids
            fields={"file":(filename,open(filelocation,'rb').read(),mimetype), "payload_json":(None,json.dumps(payload))}
        m=MultipartEncoder(fields=fields,boundary='----WebKitFormBoundary'+''.join(random.sample(string.ascii_letters+string.digits,16)))
        self.s.headers.update({"Content-Type":m.content_type})
        response = Wrapper.sendRequest(self.s, 'post', url, body=m, log=self.log)
        self.s.headers.update({"Content-Type":"application/json"})
        return response

    def reply(self, channelID, messageID, message, nonce, tts=False, embed=None, allowed_mentions={"parse":["users","roles","everyone"],"replied_user":False}, sticker_ids=None, file=None, isurl=False):
        if file == None:
            self.sendMessage(channelID, message, nonce=nonce, tts=tts, embed=embed, message_reference={"channel_id":channelID,"message_id":messageID}, allowed_mentions=allowed_mentions, sticker_ids=sticker_ids)
        else:
            self.sendFile(channelID, file, isurl=isurl, message=message, tts=tts, message_reference={"channel_id":channelID,"message_id":messageID}, sticker_ids=sticker_ids)

    def searchMessages(self,guildID,channelID,userID,mentionsUserID,has,beforeDate,afterDate,textSearch,afterNumResults, limit, extraParams): #classic discord search function, results with key "hit" are the results you searched for, afterNumResults (aka offset) is multiples of 25 and indicates after which messages (type int), filterResults defaults to False
        url = self.discord+"guilds/"+guildID+"/messages/search?"
        allqueryparams = []
        colParamNames = ["channel_id", "author_id", "mentions", "has"]
        collectionParams = [channelID, userID, mentionsUserID, has]
        for ind in range(len(collectionParams)):
            if not hasattr(collectionParams[ind], "__iter__") or isinstance(collectionParams[ind], str):
                collectionParams[ind] = [str(collectionParams[ind])]
            allqueryparams.extend([(colParamNames[ind], i) for i in collectionParams[ind]])
        if beforeDate!=None:
            allqueryparams.append(("max_id", str(beforeDate)))
        if afterDate!=None:
            allqueryparams.append(("min_id", str(afterDate)))
        if textSearch!=None:
            allqueryparams.append(("content", str(textSearch)))
        if afterNumResults!=None:
            allqueryparams.append(("offset", str(afterNumResults)))
        if limit!=None:
            allqueryparams.append(("limit", str(limit)))
        querystring = urlencode(allqueryparams)
        querystring += extraParams
        url += querystring
        return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

    def filterSearchResults(self,searchResponse): #only input is the requests response object outputted from searchMessages, returns type list
        jsonresponse = searchResponse.json()['messages']
        filteredMessages = []
        for group in jsonresponse:
            for result in group:
                if 'hit' in result:
                    filteredMessages.append(result)
        return filteredMessages

    def typingAction(self,channelID): #sends the typing action for 10 seconds (or until you change the page)
        url = self.discord+"channels/"+channelID+"/typing"
        return Wrapper.sendRequest(self.s, 'post', url, log=self.log)

    def editMessage(self,channelID,messageID,newMessage):
        url = self.discord+"channels/"+channelID+"/messages/"+messageID
        body = {"content": newMessage}
        return Wrapper.sendRequest(self.s, 'patch', url, body, log=self.log)

    def deleteMessage(self,channelID,messageID):
        url = self.discord+"channels/"+channelID+"/messages/"+messageID
        return Wrapper.sendRequest(self.s, 'delete', url, log=self.log)

    def pinMessage(self,channelID,messageID):
        url = self.discord+"channels/"+channelID+"/pins/"+messageID
        return Wrapper.sendRequest(self.s, 'put', url, log=self.log)

    def unPinMessage(self,channelID,messageID):
        url = self.discord+"channels/"+channelID+"/pins/"+messageID
        return Wrapper.sendRequest(self.s, 'delete', url, log=self.log)

    def getPins(self,channelID): #get pinned messages
        url = self.discord+"channels/"+channelID+"/pins"
        return Wrapper.sendRequest(self.s, 'get', url, log=self.log)

    def addReaction(self,channelID,messageID,emoji):
        parsedEmoji = quote_plus(emoji)
        url = self.discord+"channels/"+channelID+"/messages/"+messageID+"/reactions/"+parsedEmoji+"/%40me"
        return Wrapper.sendRequest(self.s, 'put', url, log=self.log)

    def removeReaction(self,channelID,messageID,emoji):
        parsedEmoji = quote_plus(emoji)
        url = self.discord+"channels/"+channelID+"/messages/"+messageID+"/reactions/"+parsedEmoji+"/%40me"
        return Wrapper.sendRequest(self.s, 'delete', url, log=self.log)

    #acknowledge message (mark message read)
    def ackMessage(self,channelID,messageID,ackToken):
        url = self.discord+"channels/"+channelID+"/messages/"+messageID+"/ack"
        body = {"token": ackToken}
        return Wrapper.sendRequest(self.s, 'post', url, body, log=self.log)

    #unacknowledge message (mark message unread)
    def unAckMessage(self,channelID,messageID,numMentions):
        url = self.discord+"channels/"+channelID+"/messages/"+messageID+"/ack"
        body = {"manual": True, "mention_count": numMentions}
        return Wrapper.sendRequest(self.s, 'post', url, body, log=self.log)

    def bulkAck(self, data):
        url = self.discord+"read-states/ack-bulk"
        body = {"read_states": data}
        return Wrapper.sendRequest(self.s, 'post', url, body, log=self.log)

    def getTrendingGifs(self, provider="tenor", locale="en-US", media_format="mp4"):
        url = self.discord+"gifs/trending?provider="+provider+"&locale="+locale+"&media_format="+media_format
        return Wrapper.sendRequest(self.s, 'get', url, log=self.log)
